@@ -1,5 +1,58 @@
+use core::fmt;
+
 use crate::error::Error;
 
+struct RefVisitorExpected<T>(T);
+
+impl<'de, T> fmt::Display for RefVisitorExpected<T>
+where
+    T: ReferenceVisitor<'de>,
+{
+    #[inline]
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
+        self.0.expected(f)
+    }
+}
+
+/// A visitor for data where it might be possible to borrow it without copying
+/// from the underlying [Decoder].
+///
+/// A visitor is required with [Decoder::decode_bytes] and
+/// [Decoder::decode_string] because the caller doesn't know if the encoding
+/// format is capable of producing references to the underlying data directly or
+/// if it needs to be processed.
+///
+/// By requiring a visitor we ensure that the caller has to handle both
+/// scenarios, even if one involves erroring. A type like
+/// [Cow][std::borrow::Cow] is an example of a type which can comfortably handle
+/// both.
+pub trait ReferenceVisitor<'de>: Sized {
+    /// The value being visited.
+    type Target: ?Sized;
+    /// The value produced.
+    type Ok;
+    /// The error produced.
+    type Error: Error;
+
+    /// Format an error indicating what was expected.
+    ///
+    /// Override to be more specific about the type that failed.
+    fn expected(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result;
+
+    /// Visit a string that is borrowed directly from the source data.
+    #[inline]
+    fn visit_ref(self, string: &'de Self::Target) -> Result<Self::Ok, Self::Error> {
+        self.visit(string)
+    }
+
+    /// Visit a string that is provided from the decoder in any manner possible.
+    /// Which might require additional decoding work.
+    #[inline]
+    fn visit(self, _: &Self::Target) -> Result<Self::Ok, Self::Error> {
+        Err(Self::Error::collect_from_display(RefVisitorExpected(self)))
+    }
+}
+
 /// A pack that can construct encoders.
 pub trait PackDecoder<'de> {
     /// Error type raised by this unpack.
@@ -158,10 +211,14 @@ pub trait Decoder<'de>: Sized {
     fn decode_array<const N: usize>(self) -> Result<[u8; N], Self::Error>;
 
     /// Decode a sequence of bytes whos length is encoded in the payload.
-    fn decode_bytes(self) -> Result<&'de [u8], Self::Error>;
+    fn decode_bytes<V>(self, visitor: V) -> Result<V::Ok, V::Error>
+    where
+        V: ReferenceVisitor<'de, Target = [u8], Error = Self::Error>;
 
     /// Decode a string slice from the current decoder.
-    fn decode_str(self) -> Result<&'de str, Self::Error>;
+    fn decode_string<V>(self, visitor: V) -> Result<V::Ok, V::Error>
+    where
+        V: ReferenceVisitor<'de, Target = str, Error = Self::Error>;
 
     /// Decode a boolean.
     fn decode_bool(self) -> Result<bool, Self::Error>;
